/*
 * Driver for the Atmel AC97 controller
 *
 * Copyright (C) 2005-2007 Atmel Corporation
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/mutex.h>
#include <linux/io.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/ac97_codec.h>
#include <sound/memalloc.h>

#include <asm/dma-controller.h>

#include "ac97c.h"

/* Serialize access to opened */
static DEFINE_MUTEX(opened_mutex);

struct atmel_ac97_dma_info {
	struct dma_request_cyclic req_tx;
	struct dma_request_cyclic req_rx;
	unsigned short rx_periph_id;
	unsigned short tx_periph_id;
};

struct atmel_ac97 {
	/* Serialize access to opened */
	spinlock_t lock;
	void __iomem *regs;
	struct snd_pcm_substream *playback_substream;
	struct snd_pcm_substream *capture_substream;
	struct snd_card *card;
	struct snd_pcm *pcm;
	struct snd_ac97 *ac97;
	struct snd_ac97_bus *ac97_bus;
	int opened;
	int period;
	u64 cur_format;
	unsigned int cur_rate;
	struct clk *mck;
	struct platform_device *pdev;
	struct atmel_ac97_dma_info dma;
};

#define get_chip(card) ((struct atmel_ac97 *)(card)->private_data)

#define ac97c_writel(chip, reg, val)			\
	__raw_writel((val), (chip)->regs + AC97C_##reg)
#define ac97c_readl(chip, reg)				\
	__raw_readl((chip)->regs + AC97C_##reg)

/*
 * PCM part
 */
static struct snd_pcm_hardware snd_atmel_ac97_playback_hw = {
	.info			= (SNDRV_PCM_INFO_INTERLEAVED
				  | SNDRV_PCM_INFO_MMAP
				  | SNDRV_PCM_INFO_MMAP_VALID
				  | SNDRV_PCM_INFO_BLOCK_TRANSFER
				  | SNDRV_PCM_INFO_JOINT_DUPLEX),
	.formats		= (SNDRV_PCM_FMTBIT_S16_BE
				  | SNDRV_PCM_FMTBIT_S16_LE),
	.rates			= (SNDRV_PCM_RATE_CONTINUOUS),
	.rate_min		= 4000,
	.rate_max		= 48000,
	.channels_min		= 1,
	.channels_max		= 6,
	.buffer_bytes_max	= 64*1024,
	.period_bytes_min	= 512,
	.period_bytes_max	= 4095,
	.periods_min		= 8,
	.periods_max		= 1024,
};

static struct snd_pcm_hardware snd_atmel_ac97_capture_hw = {
	.info			= (SNDRV_PCM_INFO_INTERLEAVED
				  | SNDRV_PCM_INFO_MMAP
				  | SNDRV_PCM_INFO_MMAP_VALID
				  | SNDRV_PCM_INFO_BLOCK_TRANSFER
				  | SNDRV_PCM_INFO_JOINT_DUPLEX),
	.formats		= (SNDRV_PCM_FMTBIT_S16_BE
				  | SNDRV_PCM_FMTBIT_S16_LE),
	.rates			= (SNDRV_PCM_RATE_CONTINUOUS),
	.rate_min		= 4000,
	.rate_max		= 48000,
	.channels_min		= 1,
	.channels_max		= 2,
	.buffer_bytes_max	= 64*1024,
	.period_bytes_min	= 512,
	.period_bytes_max	= 4095,
	.periods_min		= 8,
	.periods_max		= 1024,
};

/*
 * PCM functions
 */
static int
snd_atmel_ac97_playback_open(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	mutex_lock(&opened_mutex);
	chip->opened++;
	runtime->hw = snd_atmel_ac97_playback_hw;
	if (chip->cur_rate) {
		runtime->hw.rate_min = chip->cur_rate;
		runtime->hw.rate_max = chip->cur_rate;
	}
	if (chip->cur_format)
		runtime->hw.formats = (1ULL << chip->cur_format);
	mutex_unlock(&opened_mutex);
	chip->playback_substream = substream;
	chip->period = 0;
	return 0;
}

static int
snd_atmel_ac97_capture_open(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	mutex_lock(&opened_mutex);
	chip->opened++;
	runtime->hw = snd_atmel_ac97_capture_hw;
	if (chip->cur_rate) {
		runtime->hw.rate_min = chip->cur_rate;
		runtime->hw.rate_max = chip->cur_rate;
	}
	if (chip->cur_format)
		runtime->hw.formats = (1ULL << chip->cur_format);
	mutex_unlock(&opened_mutex);
	chip->capture_substream = substream;
	chip->period = 0;
	return 0;
}

static int snd_atmel_ac97_playback_close(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	mutex_lock(&opened_mutex);
	chip->opened--;
	if (!chip->opened) {
		chip->cur_rate = 0;
		chip->cur_format = 0;
	}
	mutex_unlock(&opened_mutex);
	return 0;
}

static int snd_atmel_ac97_capture_close(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	mutex_lock(&opened_mutex);
	chip->opened--;
	if (!chip->opened) {
		chip->cur_rate = 0;
		chip->cur_format = 0;
	}
	mutex_unlock(&opened_mutex);
	return 0;
}

static int
snd_atmel_ac97_playback_hw_params(struct snd_pcm_substream *substream,
		struct snd_pcm_hw_params *hw_params)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	int err;

	err = snd_pcm_lib_malloc_pages(substream,
					params_buffer_bytes(hw_params));
	if (err < 0)
		return err;

	/* Set restrictions to params */
	mutex_lock(&opened_mutex);
	chip->cur_rate = params_rate(hw_params);
	chip->cur_format = params_format(hw_params);
	mutex_unlock(&opened_mutex);

	return 0;
}

static int
snd_atmel_ac97_capture_hw_params(struct snd_pcm_substream *substream,
		struct snd_pcm_hw_params *hw_params)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	int err;

	err = snd_pcm_lib_malloc_pages(substream,
					params_buffer_bytes(hw_params));
	if (err < 0)
		return err;

	/* Set restrictions to params */
	mutex_lock(&opened_mutex);
	chip->cur_rate = params_rate(hw_params);
	chip->cur_format = params_format(hw_params);
	mutex_unlock(&opened_mutex);

	return 0;
}

static int snd_atmel_ac97_playback_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

static int snd_atmel_ac97_capture_hw_free(struct snd_pcm_substream *substream)
{

	return snd_pcm_lib_free_pages(substream);
}

static int snd_atmel_ac97_playback_prepare(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct platform_device *pdev = chip->pdev;
	struct snd_pcm_runtime *runtime = substream->runtime;
	int block_size = frames_to_bytes(runtime, runtime->period_size);
	unsigned long word = 0;
	unsigned long buffer_size = 0;

	dma_sync_single_for_device(&pdev->dev, runtime->dma_addr,
			block_size * 2, DMA_TO_DEVICE);

	/* Assign slots to channels */
	switch (substream->runtime->channels) {
	case 1:
		word |= AC97C_CH_ASSIGN(PCM_LEFT, A);
		break;
	case 2:
		/* Assign Left and Right slot to Channel A */
		word |= AC97C_CH_ASSIGN(PCM_LEFT, A)
			| AC97C_CH_ASSIGN(PCM_RIGHT, A);
		break;
	default:
		/* TODO: support more than two channels */
		return -EINVAL;
		break;
	}
	ac97c_writel(chip, OCA, word);

	/* Configure sample format and size */
	word = AC97C_CMR_PDCEN | AC97C_CMR_SIZE_16;

	switch (runtime->format) {
	case SNDRV_PCM_FORMAT_S16_LE:
		word |= AC97C_CMR_CEM_LITTLE;
		break;
	case SNDRV_PCM_FORMAT_S16_BE: /* fall through */
	default:
		word &= ~AC97C_CMR_CEM_LITTLE;
		break;
	}

	ac97c_writel(chip, CAMR, word);

	/* Set variable rate if needed */
	if (runtime->rate != 48000) {
		word = ac97c_readl(chip, MR);
		word |= AC97C_MR_VRA;
		ac97c_writel(chip, MR, word);
	} else {
		/* Clear Variable Rate Bit */
		word = ac97c_readl(chip, MR);
		word &= ~AC97C_MR_VRA;
		ac97c_writel(chip, MR, word);
	}

	/* Set rate */
	snd_ac97_set_rate(chip->ac97, AC97_PCM_FRONT_DAC_RATE, runtime->rate);

	buffer_size = frames_to_bytes(runtime, runtime->period_size) *
		runtime->periods;

	chip->dma.req_tx.buffer_size = buffer_size;
	chip->dma.req_tx.periods = runtime->periods;

	BUG_ON(chip->dma.req_tx.buffer_size !=
			(chip->dma.req_tx.periods *
			 frames_to_bytes(runtime, runtime->period_size)));

	chip->dma.req_tx.buffer_start = runtime->dma_addr;
	chip->dma.req_tx.data_reg = (dma_addr_t)(chip->regs + AC97C_CATHR + 2);
	chip->dma.req_tx.periph_id = chip->dma.tx_periph_id;
	chip->dma.req_tx.direction = DMA_DIR_MEM_TO_PERIPH;
	chip->dma.req_tx.width = DMA_WIDTH_16BIT;
	chip->dma.req_tx.dev_id = chip;

	return 0;
}

static int snd_atmel_ac97_capture_prepare(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct platform_device *pdev = chip->pdev;
	struct snd_pcm_runtime *runtime = substream->runtime;
	int block_size = frames_to_bytes(runtime, runtime->period_size);
	unsigned long word = 0;
	unsigned long buffer_size = 0;

	dma_sync_single_for_device(&pdev->dev, runtime->dma_addr,
			block_size * 2, DMA_FROM_DEVICE);

	/* Assign slots to channels */
	switch (substream->runtime->channels) {
	case 1:
		word |= AC97C_CH_ASSIGN(PCM_LEFT, A);
		break;
	case 2:
		/* Assign Left and Right slot to Channel A */
		word |= AC97C_CH_ASSIGN(PCM_LEFT, A)
			| AC97C_CH_ASSIGN(PCM_RIGHT, A);
		break;
	default:
		/* TODO: support more than two channels */
		return -EINVAL;
		break;
	}
	ac97c_writel(chip, ICA, word);

	/* Configure sample format and size */
	word = AC97C_CMR_PDCEN | AC97C_CMR_SIZE_16;

	switch (runtime->format) {
	case SNDRV_PCM_FORMAT_S16_LE:
		word |= AC97C_CMR_CEM_LITTLE;
		break;
	case SNDRV_PCM_FORMAT_S16_BE:
	default:
		word &= ~(AC97C_CMR_CEM_LITTLE);
		break;
	}

	ac97c_writel(chip, CAMR, word);

	/* Set variable rate if needed */
	if (runtime->rate != 48000) {
		word = ac97c_readl(chip, MR);
		word |= AC97C_MR_VRA;
		ac97c_writel(chip, MR, word);
	} else {
		/* Clear Variable Rate Bit */
		word = ac97c_readl(chip, MR);
		word &= ~(AC97C_MR_VRA);
		ac97c_writel(chip, MR, word);
	}

	/* Set rate */
	snd_ac97_set_rate(chip->ac97, AC97_PCM_LR_ADC_RATE, runtime->rate);

	buffer_size = frames_to_bytes(runtime, runtime->period_size) *
		runtime->periods;

	chip->dma.req_rx.buffer_size = buffer_size;
	chip->dma.req_rx.periods = runtime->periods;

	BUG_ON(chip->dma.req_rx.buffer_size !=
			(chip->dma.req_rx.periods *
			 frames_to_bytes(runtime, runtime->period_size)));

	chip->dma.req_rx.buffer_start = runtime->dma_addr;
	chip->dma.req_rx.data_reg = (dma_addr_t)(chip->regs + AC97C_CARHR + 2);
	chip->dma.req_rx.periph_id = chip->dma.rx_periph_id;
	chip->dma.req_rx.direction = DMA_DIR_PERIPH_TO_MEM;
	chip->dma.req_rx.width = DMA_WIDTH_16BIT;
	chip->dma.req_rx.dev_id = chip;

	return 0;
}

	static int
snd_atmel_ac97_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	unsigned long camr;
	int flags, err = 0;

	spin_lock_irqsave(&chip->lock, flags);
	camr = ac97c_readl(chip, CAMR);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		err = dma_prepare_request_cyclic(chip->dma.req_tx.req.dmac,
				&chip->dma.req_tx);
		dma_start_request(chip->dma.req_tx.req.dmac,
				chip->dma.req_tx.req.channel);
		camr |= AC97C_CMR_CENA;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		err = dma_stop_request(chip->dma.req_tx.req.dmac,
				chip->dma.req_tx.req.channel);
		if (chip->opened <= 1)
			camr &= ~AC97C_CMR_CENA;
		break;
	default:
		err = -EINVAL;
		break;
	}

	ac97c_writel(chip, CAMR, camr);

	spin_unlock_irqrestore(&chip->lock, flags);
	return err;
}

	static int
snd_atmel_ac97_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	unsigned long camr;
	int flags, err = 0;

	spin_lock_irqsave(&chip->lock, flags);
	camr = ac97c_readl(chip, CAMR);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		err = dma_prepare_request_cyclic(chip->dma.req_rx.req.dmac,
				&chip->dma.req_rx);
		dma_start_request(chip->dma.req_rx.req.dmac,
				chip->dma.req_rx.req.channel);
		camr |= AC97C_CMR_CENA;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		err = dma_stop_request(chip->dma.req_rx.req.dmac,
				chip->dma.req_rx.req.channel);
		mutex_lock(&opened_mutex);
		if (chip->opened <= 1)
			camr &= ~AC97C_CMR_CENA;
		mutex_unlock(&opened_mutex);
		break;
	default:
		err = -EINVAL;
		break;
	}

	ac97c_writel(chip, CAMR, camr);

	spin_unlock_irqrestore(&chip->lock, flags);
	return err;
}

	static snd_pcm_uframes_t
snd_atmel_ac97_playback_pointer(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t pos;
	unsigned long bytes;

	bytes = (dma_get_current_pos
			(chip->dma.req_tx.req.dmac,
			 chip->dma.req_tx.req.channel) - runtime->dma_addr);
	pos = bytes_to_frames(runtime, bytes);
	if (pos >= runtime->buffer_size)
		pos -= runtime->buffer_size;

	return pos;
}

	static snd_pcm_uframes_t
snd_atmel_ac97_capture_pointer(struct snd_pcm_substream *substream)
{
	struct atmel_ac97 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t pos;
	unsigned long bytes;

	bytes = (dma_get_current_pos
			(chip->dma.req_rx.req.dmac,
			 chip->dma.req_rx.req.channel)
			- runtime->dma_addr);
	pos = bytes_to_frames(runtime, bytes);
	if (pos >= runtime->buffer_size)
		pos -= runtime->buffer_size;


	return pos;
}

static struct snd_pcm_ops atmel_ac97_playback_ops = {
	.open		= snd_atmel_ac97_playback_open,
	.close		= snd_atmel_ac97_playback_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= snd_atmel_ac97_playback_hw_params,
	.hw_free	= snd_atmel_ac97_playback_hw_free,
	.prepare	= snd_atmel_ac97_playback_prepare,
	.trigger	= snd_atmel_ac97_playback_trigger,
	.pointer	= snd_atmel_ac97_playback_pointer,
};

static struct snd_pcm_ops atmel_ac97_capture_ops = {
	.open		= snd_atmel_ac97_capture_open,
	.close		= snd_atmel_ac97_capture_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= snd_atmel_ac97_capture_hw_params,
	.hw_free	= snd_atmel_ac97_capture_hw_free,
	.prepare	= snd_atmel_ac97_capture_prepare,
	.trigger	= snd_atmel_ac97_capture_trigger,
	.pointer	= snd_atmel_ac97_capture_pointer,
};

static struct ac97_pcm atmel_ac97_pcm_defs[] __devinitdata = {
	/* Playback */
	{
		.exclusive = 1,
		.r = { {
			.slots = ((1 << AC97_SLOT_PCM_LEFT)
					| (1 << AC97_SLOT_PCM_RIGHT)
					| (1 << AC97_SLOT_PCM_CENTER)
					| (1 << AC97_SLOT_PCM_SLEFT)
					| (1 << AC97_SLOT_PCM_SRIGHT)
					| (1 << AC97_SLOT_LFE)),
		} }
	},
	/* PCM in */
	{
		.stream = 1,
		.exclusive = 1,
		.r = { {
			.slots = ((1 << AC97_SLOT_PCM_LEFT)
					| (1 << AC97_SLOT_PCM_RIGHT)),
		} }
	},
	/* Mic in */
	{
		.stream = 1,
		.exclusive = 1,
		.r = { {
			.slots = (1<<AC97_SLOT_MIC),
		} }
	},
};

static int __devinit snd_atmel_ac97_pcm_new(struct atmel_ac97 *chip)
{
	struct snd_pcm *pcm;
	int err;

	err = snd_ac97_pcm_assign(chip->ac97_bus,
			ARRAY_SIZE(atmel_ac97_pcm_defs),
			atmel_ac97_pcm_defs);
	if (err)
		return err;

	err = snd_pcm_new(chip->card, "Atmel-AC97", 0, 1, 1, &pcm);
	if (err)
		return err;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
			&atmel_ac97_playback_ops);

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&atmel_ac97_capture_ops);

	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
			&chip->pdev->dev,
			128 * 1024, 128 * 1024);

	pcm->private_data = chip;
	pcm->info_flags = 0;
	strcpy(pcm->name, "Atmel-AC97");
	chip->pcm = pcm;

	return 0;
}

/*
 * Mixer part.
 */
static int snd_atmel_ac97_mixer_new(struct atmel_ac97 *chip)
{
	int err;
	struct snd_ac97_template template;

	memset(&template, 0, sizeof(template));
	template.private_data = chip;
	err = snd_ac97_mixer(chip->ac97_bus, &template, &chip->ac97);

	return err;
}

static void atmel_ac97_error(struct dma_request *_req)
{
	struct dma_request_cyclic *req = to_dma_request_cyclic(_req);
	struct atmel_ac97 *chip = req->dev_id;

	dev_dbg(&chip->pdev->dev, "DMA Controller error, channel %d\n",
			req->req.channel);
}

static void atmel_ac97_block_complete(struct dma_request *_req)
{
	struct dma_request_cyclic *req = to_dma_request_cyclic(_req);
	struct atmel_ac97 *chip = req->dev_id;
	if (req->periph_id == chip->dma.tx_periph_id)
		snd_pcm_period_elapsed(chip->playback_substream);
	else
		snd_pcm_period_elapsed(chip->capture_substream);
}

/*
 * Codec part.
 */
static void snd_atmel_ac97_write(struct snd_ac97 *ac97, unsigned short reg,
		unsigned short val)
{
	struct atmel_ac97 *chip = get_chip(ac97);
	unsigned long word;
	int timeout = 40;

	word = (reg & 0x7f) << 16 | val;

	do {
		if (ac97c_readl(chip, COSR) & AC97C_CSR_TXRDY) {
			ac97c_writel(chip, COTHR, word);
			return;
		}
		udelay(1);
	} while (--timeout);

	dev_dbg(&chip->pdev->dev, "codec write timeout\n");
}

static unsigned short snd_atmel_ac97_read(struct snd_ac97 *ac97,
		unsigned short reg)
{
	struct atmel_ac97 *chip = get_chip(ac97);
	unsigned long word;
	int timeout = 40;
	int write = 10;

	word = (0x80 | (reg & 0x7f)) << 16;

	if ((ac97c_readl(chip, COSR) & AC97C_CSR_RXRDY) != 0)
		ac97c_readl(chip, CORHR);

retry_write:
	timeout = 40;

	do {
		if ((ac97c_readl(chip, COSR) & AC97C_CSR_TXRDY) != 0) {
			ac97c_writel(chip, COTHR, word);
			goto read_reg;
		}
		mdelay(10);
	} while (--timeout);

	if (!--write)
		goto timed_out;
	goto retry_write;

read_reg:
	do {
		if ((ac97c_readl(chip, COSR) & AC97C_CSR_RXRDY) != 0) {
			unsigned short val = ac97c_readl(chip, CORHR);
			return val;
		}
		mdelay(10);
	} while (--timeout);

	if (!--write)
		goto timed_out;
	goto retry_write;

timed_out:
	dev_dbg(&chip->pdev->dev, "codec read timeout\n");
	return 0xffff;
}

static void snd_atmel_ac97_reset(struct atmel_ac97 *chip)
{
	ac97c_writel(chip, MR, AC97C_MR_WRST);
	mdelay(1);
	ac97c_writel(chip, MR, AC97C_MR_ENA);
}

static void snd_atmel_ac97_destroy(struct snd_card *card)
{
	struct atmel_ac97 *chip = get_chip(card);

	if (chip->regs)
		iounmap(chip->regs);

	if (chip->mck) {
		clk_disable(chip->mck);
		clk_put(chip->mck);
	}

	if (chip->dma.req_tx.req.dmac) {
		dma_release_channel(chip->dma.req_tx.req.dmac,
				chip->dma.req_tx.req.channel);
	}
	if (chip->dma.req_rx.req.dmac) {
		dma_release_channel(chip->dma.req_rx.req.dmac,
				chip->dma.req_rx.req.channel);
	}
}

static int __devinit snd_atmel_ac97_create(struct snd_card *card,
		struct platform_device *pdev)
{
	static struct snd_ac97_bus_ops ops = {
		.write	= snd_atmel_ac97_write,
		.read	= snd_atmel_ac97_read,
	};
	struct atmel_ac97 *chip = get_chip(card);
	struct resource *regs;
	struct clk *mck;
	int err;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	mck = clk_get(&pdev->dev, "pclk");
	if (IS_ERR(mck))
		return PTR_ERR(mck);
	clk_enable(mck);
	chip->mck = mck;

	card->private_free = snd_atmel_ac97_destroy;

	spin_lock_init(&chip->lock);
	chip->card = card;
	chip->pdev = pdev;

	chip->regs = ioremap(regs->start, regs->end - regs->start + 1);
	if (!chip->regs)
		return -ENOMEM;

	snd_card_set_dev(card, &pdev->dev);

	err = snd_ac97_bus(card, 0, &ops, chip, &chip->ac97_bus);

	return err;
}

static int __devinit snd_atmel_ac97_probe(struct platform_device *pdev)
{
	static int dev;
	struct snd_card *card;
	struct atmel_ac97 *chip;
	int err;
	int ch;

	mutex_init(&opened_mutex);

	err = -ENOMEM;
	card = snd_card_new(SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1,
			THIS_MODULE, sizeof(struct atmel_ac97));
	if (!card)
		goto out;
	chip = get_chip(card);

	err = snd_atmel_ac97_create(card, pdev);
	if (err)
		goto out_free_card;

	snd_atmel_ac97_reset(chip);

	err = snd_atmel_ac97_mixer_new(chip);
	if (err)
		goto out_free_card;

	err = snd_atmel_ac97_pcm_new(chip);
	if (err)
		goto out_free_card;

	/* TODO: Get this information from the platform device */
	chip->dma.req_tx.req.dmac = find_dma_controller(0);
	if (!chip->dma.req_tx.req.dmac) {
		dev_dbg(&chip->pdev->dev, "DMA controller for TX missing\n");
		err = -ENODEV;
		goto out_free_card;
	}
	chip->dma.req_rx.req.dmac = find_dma_controller(0);
	if (!chip->dma.req_rx.req.dmac) {
		dev_dbg(&chip->pdev->dev, "DMA controller for RX missing\n");
		err = -ENODEV;
		goto out_free_card;
	}

	chip->dma.rx_periph_id = 3;
	chip->dma.tx_periph_id = 4;

	ch = dma_alloc_channel(chip->dma.req_tx.req.dmac);
	if (ch < 0) {
		dev_dbg(&chip->pdev->dev,
				"could not allocate TX DMA channel\n");
		err = ch;
		goto out_free_card;
	}
	chip->dma.req_tx.req.channel = ch;
	chip->dma.req_tx.width = DMA_WIDTH_16BIT;
	chip->dma.req_tx.req.block_complete = atmel_ac97_block_complete;
	chip->dma.req_tx.req.error = atmel_ac97_error;

	ch = dma_alloc_channel(chip->dma.req_rx.req.dmac);
	if (ch < 0) {
		dev_dbg(&chip->pdev->dev,
				"could not allocate RX DMA channel\n");
		err = ch;
		goto out_free_card;
	}
	chip->dma.req_rx.req.channel = ch;
	chip->dma.req_rx.width = DMA_WIDTH_16BIT;
	chip->dma.req_rx.req.block_complete = atmel_ac97_block_complete;
	chip->dma.req_rx.req.error = atmel_ac97_error;

	strcpy(card->driver, "atmel_ac97c");
	strcpy(card->shortname, "atmel_ac97c");
	sprintf(card->longname, "Atmel AVR32 AC97 controller");

	err = snd_card_register(card);
	if (err)
		goto out_free_card;

	platform_set_drvdata(pdev, card);
	dev++;

	dev_info(&pdev->dev, "Atmel AVR32 AC97 controller at 0x%p\n",
			chip->regs);

	return 0;

out_free_card:
	snd_card_free(card);
out:
	return err;
}

#ifdef CONFIG_PM
	static int
snd_atmel_ac97_suspend(struct platform_device *pdev, pm_message_t msg)
{
	struct snd_card *card = platform_get_drvdata(pdev);
	struct atmel_ac97 *chip = card->private_data;

	clk_disable(chip->mck);

	return 0;
}

static int snd_atmel_ac97_resume(struct platform_device *pdev)
{
	struct snd_card *card = dev_get_drvdata(pdev);
	struct atmel_ac97 *chip = card->private_data;

	clk_enable(chip->mck);

	return 0;
}
#else
#define snd_atmel_ac97_suspend NULL
#define snd_atmel_ac97_resume NULL
#endif

static int __devexit snd_atmel_ac97_remove(struct platform_device *pdev)
{
	struct snd_card *card = platform_get_drvdata(pdev);

	snd_card_free(card);
	platform_set_drvdata(pdev, NULL);
	return 0;
}

static struct platform_driver atmel_ac97_driver = {
	.remove		= __devexit_p(snd_atmel_ac97_remove),
	.driver		= {
		.name	= "atmel_ac97c",
	},
	.suspend	= snd_atmel_ac97_suspend,
	.resume		= snd_atmel_ac97_resume,
};

static int __init atmel_ac97_init(void)
{
	return platform_driver_probe(&atmel_ac97_driver,
			snd_atmel_ac97_probe);
}
module_init(atmel_ac97_init);

static void __exit atmel_ac97_exit(void)
{
	platform_driver_unregister(&atmel_ac97_driver);
}
module_exit(atmel_ac97_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for Atmel AC97 Controller");
MODULE_AUTHOR("Haavard Skinnemoen <hskinnemoen@atmel.com>");
